# coding: utf-8
# create by tongshiwei on 2017/12/2
# from __future__ import absolute_import

# py2 aborted
# def unistr(x):
#     return x

# py2 aborted
# def tostr(x):
#     return str(x)

# string_types = basestring,
# integer_types = (int, long)
# unistr = lambda x: x.decode('utf-8')
# tostr = lambda x: unicode(x)
