# coding: utf-8
# 2021/8/5 @ tongshiwei

from .utils import getF, copy_net, array_equal
from .sequence import get_begin_state, mask_sequence_variable_length, format_sequence
from .embedding import set_embedding_weight
