# coding: utf-8
# 2020/4/13 @ tongshiwei


from .classification import classification_report
from .regression import regression_report
from .ranking import ranking_report
from .utils import POrderedDict
